/* 
 * Copyright (C) 2000-2001 plitsch-platsch
 * 
 * xine_dvd_plugin is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * xine_dvd_plugin is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *
 */

#include <dlfcn.h>
#include <stdio.h>
#include <fcntl.h>
#if defined(__NetBSD__) || defined(__OpenBSD__) || defined(__FreeBSD__)
# include <sys/dvdio.h>
# include <sys/cdio.h> /* CDIOCALLOW etc... */
#elif defined(__linux__)
# include <linux/cdrom.h>
#else
# error "Need the DVD ioctls"
#endif
#include <sys/ioctl.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <inttypes.h>

#include "input_plugin.h"
#include "monitor.h"
#include "dvd_udf.h"
#include "css.h"
#include "ifo.h"
#include "bswap.h"

#define DVD     "/dev/dvd"
#define RDVD    "/dev/rdvd"

#define MAX_DIR_ENTRIES 1024
  
/* 
 * Define this and input_plugin_get_autoplay_list() 
 * will return all of titles/chapters
 */
//#define ALL_TITLES 1


/*
 * global Variables:
 */
typedef struct {
  input_plugin_t     input_plugin;
  
  char              *mrl;
  config_values_t   *config;

  int                bEncrypted;
  char               key_disk [2048];
  char               key_title [5];
  off_t              file_size, file_size_left;
  off_t              file_lbstart, file_lbcur;
  int                dvd_fd, raw_fd;
  int                gVTSMinor, gVTSMajor;
  
  /*
   * ifo stuff
   */
  ifo_t               *ifo;
  off_t                title_start;
  pgc_t               *pgci;          
  ifo_pgci_caddr_t    *cpli;          
  ifo_pgc_cpos_t      *cpos;          
  int32_t              cpos_index;    
  ifo_caddr_t         *caddr;         
  int32_t              caddr_index;    
  uint32_t             caddr_index_max;
  uint32_t             cellp_pos;
  uint32_t             cell_end; 
  off_t                byte_pos;
  uint8_t             *clink;
  int                  chapter, title, to_chapter;
  uint32_t             lb_end; /* lb to stop at (==start lb of next chapter) */
  ifo_ptt_t           *ptt;

  /*
   * udf dir function 
   */
  char                *filelist[MAX_DIR_ENTRIES];
  int                  num_filelist;
  char                *filelist2[MAX_DIR_ENTRIES];
  int                  num_filelist2;

  mrl_t               *mrls[MAX_DIR_ENTRIES];
  int                  mrls_allocated_entries;
} dvd_input_plugin_t;

#define MAX(a,b) ( (a) > (b) ? (a) : (b) )
#define MIN(a,b) ( (a) < (b) ? (a) : (b) )

static void reset_dvd_drive (void) {
  int fd;
#ifndef __FreeBSD__
  dvd_authinfo ai;
#else
  struct dvd_authinfo ai;
#endif

  int i;

  /*
   * reset drive 
   */

  fd = open(DVD, O_RDONLY | O_NONBLOCK);

  if (fd < 0) {
    printf ("input_d4d: reset_dvd_drive: unable to open dvd drive (%s): %s\n", DVD, strerror(errno));
    return ;
  }

  for (i = 0; i < 4; i++) {
    memset(&ai, 0, sizeof(ai));
#ifndef __FreeBSD__
    ai.type = DVD_INVALIDATE_AGID;
    ai.lsa.agid = i;
    ioctl(fd, DVD_AUTH, &ai);
#else
    ai.format = DVD_INVALIDATE_AGID;
    ai.agid = i;
    ioctl(fd, DVDIOCREPORTKEY, &ai);
#endif
  }

  close (fd); 
}

static int open_drive (dvd_input_plugin_t *this) {
  
  this->dvd_fd = open(DVD, O_RDONLY | O_NONBLOCK);

  if (this->dvd_fd < 0) {
    printf ("input_d4d: unable to open dvd drive (%s): %s\n", DVD, strerror(errno));
    return -1;
  }

  this->raw_fd = open(RDVD, O_RDONLY | O_NONBLOCK);
  if (this->raw_fd < 0) {
    this->raw_fd = this->dvd_fd;
  }

  return this->raw_fd;
}

static void close_drive (dvd_input_plugin_t *this) {

  if (this->dvd_fd<0)
    return;

  close (this->dvd_fd);
  if (this->raw_fd != this->dvd_fd)
    close (this->raw_fd);

  this->dvd_fd = -1;
}

static void unlockDrive (dvd_input_plugin_t *this, off_t lbnum) {

  lseek (this->dvd_fd, lbnum * (off_t) DVD_VIDEO_LB_LEN, SEEK_SET) ;

  /* printf ("seek to %d: %d\n",lbnum, ret); */
  
  CSSAuthDisc (this->dvd_fd, this->key_disk);
  
  close_drive (this);
  
  open_drive (this);
  
  lseek (this->dvd_fd, lbnum * (off_t) DVD_VIDEO_LB_LEN, SEEK_SET) ;

  if (CSSAuthTitle (this->dvd_fd, this->key_title, lbnum)>=0) {
    
/*    if (CSSDecryptTitleKey (this->key_title, this->key_disk) >= 0) { */
    if (CSSDecryptTitleKey2 (this->dvd_fd, this->key_title, lbnum * (off_t) DVD_VIDEO_LB_LEN) >= 0) {
      
      this->bEncrypted = 1;
    } else {
      perror ("Decrypting title key");
      
      this->bEncrypted = 0;
    }
    
  } else {
    this->bEncrypted = 0;
    /*perror ("Authenticate title");*/
  }
}

/*
 * try to open dvd, authenticate and prepare to read/decrypt >filename<
 *
 * returns lbnum on success, 0 otherwise
 */
static int openDVDFile (dvd_input_plugin_t *this, 
			char *filename, off_t *size) {

#ifndef __FreeBSD__
  dvd_struct         dvd;
#else
  struct dvd_struct  dvd;
#endif

  char               str[256];
  off_t              lbnum;

  printf ("input_d4d : openDVDFile >%s<\n", filename);

  reset_dvd_drive ();

  
  if (open_drive(this) < 0) 
    return 0;

  /*
   * encrypted ?
   */

#ifndef __FreeBSD__
  dvd.copyright.type = DVD_STRUCT_COPYRIGHT;
  dvd.copyright.layer_num=0;

  if (ioctl (this->dvd_fd, DVD_READ_STRUCT, &dvd) < 0) {
    perror ("DVD_READ_STRUCT");
    return 0;
  }

  this->bEncrypted = (dvd.copyright.cpst != 0) ;

#else
  dvd.format    = DVD_STRUCT_COPYRIGHT;
  dvd.layer_num = 0;
  
  if (ioctl(this->dvd_fd, DVDIOCREADSTRUCTURE, &dvd) < 0) {
    fprintf (stderr, "Could not read Copyright Struct\n");
    return 0;
  }

  this->bEncrypted = (dvd.cpst != 0);
#endif

  close_drive (this);

  if (this->bEncrypted)
    printf ("encrypted\n");
  else
    printf ("unencrypted\n");

  open_drive (this);

  sprintf (str, "/VIDEO_TS/%s", filename);

  printf ("UDFFindFile %s\n", str);
    
  if ((lbnum = UDFFindFile(this->dvd_fd, str, size))) {

    unlockDrive (this, lbnum);

    lseek (this->raw_fd, lbnum * (off_t) DVD_VIDEO_LB_LEN, SEEK_SET) ;

  } else {
    printf ("input_d4d: cannot open file >%s<\n", filename);

    close_drive (this);

    return 0;
  }

  return lbnum;
}

static int _dvd_next_cell (dvd_input_plugin_t *this) {
  
  int old_caddr_index = this->caddr_index;

  /*
  printf( "looking for cell %x:%x (%x/%x)\n", 
	  this->cpos[this->cpos_index].vob_id, 
	  this->cpos[this->cpos_index].cell_id, 
	  this->cpos_index, this->caddr_index_max);
  */

  if (this->cpos_index > this->caddr_index_max) {
    printf("input_d4d: cpos_index >caddr_index_max => finish\n");
    return -1;
  }

  while (++this->caddr_index < this->caddr_index_max) {

    if ((this->cpos[this->cpos_index].vob_id == this->caddr[this->caddr_index].vob_id)
	&& (this->cpos[this->cpos_index].cell_id == this->caddr[this->caddr_index].cell_id)) {
      goto cell_found;
    }
  }

  this->caddr_index = old_caddr_index;

  return -1;

cell_found:

  /* for ILVUs which don't end at cell boundary */
  this->cellp_pos = MAX (be2me_32 (this->caddr[this->caddr_index].start), 
		   be2me_32 (this->cpli[this->cpos_index].vobu_start));
  this->cell_end = MIN (be2me_32 (this->caddr[this->caddr_index].end), 
		  be2me_32 (this->cpli[this->cpos_index].vobu_last_end));

  /* printf("|%x %x|", cellp_pos, cell_end); */
  return 0;
}

static int get_num_ifos (dvd_input_plugin_t *this) {

  int      found_num, i, num, sub_num;

  UDFListDir (this->dvd_fd, "/VIDEO_TS", MAX_DIR_ENTRIES, 
	      this->filelist, &this->num_filelist);

  if(this->num_filelist < 0)
    return 0;
  
  found_num = 0;
  for (i=0; i<this->num_filelist; i++) {
    if (sscanf (this->filelist[i], "VTS_%d_%d.VOB", &num, &sub_num)==2) {
      if (num>found_num)
	found_num = num;
    }
  }

  return found_num;
}


static int find_movie_ifo (dvd_input_plugin_t *this) {

  int      found_num, found_sub, num, subnum, i;

  UDFListDir (this->dvd_fd, "/VIDEO_TS", MAX_DIR_ENTRIES, 
	      this->filelist, &this->num_filelist);

  if(this->num_filelist < 0)
    return -1;
  
  this->title_start = 0;
  found_num = -1; found_sub = -1; 
  for (i=0; i<this->num_filelist; i++) {
    if (sscanf (this->filelist[i], "VTS_%d_%d.VOB", &num, &subnum)==2) {
      if (subnum > found_sub) {
	found_sub = subnum;
	found_num = num;
      }
    }
  }

  return found_num;
}

static int parse_ifo (dvd_input_plugin_t *this, int ifo_num) {

  off_t    size;

  char str[1024];
  sprintf (str, "/VIDEO_TS/VTS_%02d_0.IFO", ifo_num);
  
  this->title_start = UDFFindFile (this->dvd_fd, str, &size);
  if (!this->title_start) {
    printf ("input_d4d: failed to find %s\n", str);
    return 0;
  }
    
  this->ifo = ifoOpen (this->dvd_fd, this->title_start * 2048);
  
  if (!this->ifo) {
    printf ("input_d4d: ifoOpen 2 failed\n");
    return 0;
  }

  this->caddr = (ifo_caddr_t*)ifoGetCellAddr (this->ifo);
  this->title_start += (off_t) ifoGetVOBStart (this->ifo);

  this->ptt = ifoGetPTT (this->ifo);

  unlockDrive (this, this->title_start);

  return 1;
}

static int openDVDChapter (dvd_input_plugin_t *this, int ifo_num,
			   int title_num, int chapter, int to_chapter) {

  ifo_ptt_sub_t *title;

  this->title      = title_num;
  this->chapter    = chapter;
  this->to_chapter = to_chapter;

  reset_dvd_drive ();

  if (open_drive(this)<0) 
    return 0;
  
  /*
   * parse IFO files
   */

  if (!parse_ifo (this, ifo_num))
    return 0;

  title = this->ptt->title + title_num;

  /*
   * use cell link to find first cell of selected chapter 
   */

  this->pgci = ifoGetPGCI (this->ifo, IFO_TITLE, title_num);
  
  if (!this->pgci) {
    printf ("error get PGCI\n");
    return 0;
  }

  this->clink = (uint8_t*)ifoGetProgramMap (this->pgci);

  this->caddr_index_max = ifoGetCellAddrNum (this->ifo, IFO_TITLE);

  /* Cell Position */
  if (!(this->cpos = ifoGetCellPos (this->pgci))) {
    printf ("error ifoGetCellPos");
    return 0;
  }

  /* Cell Play Information */
  if (!(this->cpli = ifoGetCellPlayInfo (this->pgci))) {
    printf("error ifoGetCellPlayInfo");
    return 0;
  }

  /*
   * find out start lb of next chapter (if there is one)
   */

  if (to_chapter < (title->num-1)) {
    this->cpos_index     = this->clink[to_chapter+1]-1;
    this->caddr_index    = -1;
    this->cellp_pos      = 0;
    this->cell_end       = 0;

    if (++this->cellp_pos > this->cell_end) {
      if (_dvd_next_cell (this) < 0) {
	this->cpos_index++;
	_dvd_next_cell (this);
      }
    }

    this->lb_end = this->title_start + this->cellp_pos;

  } else {
    this->lb_end = 0;
  }

  /*
   * calculate length of input
   * (I know this is pathetic, but it works! ;> )
   */

  this->cpos_index     = this->clink[chapter]-1;
  this->caddr_index    = -1;
  this->cellp_pos      = 0;
  this->cell_end       = 0;
  
  this->file_size = 0;

  while (1) {
    if ((this->cellp_pos + this->title_start) == this->lb_end) /* end reached ? */
      break;
    if (++this->cellp_pos > this->cell_end) {       /* next cell */
      if (_dvd_next_cell (this) < 0) {
	this->cpos_index++;
	if (_dvd_next_cell (this) < 0)
	  break;
      }
    }
    this->file_size += DVD_VIDEO_LB_LEN;
  }

  this->file_size_left = this->file_size;

  printf ("input length : %lld\n", this->file_size);

  /* reset values after length calculation */

  this->cpos_index     = this->clink[chapter]-1;

  this->caddr_index = -1;
  this->cellp_pos = 0;
  this->cell_end = 0;
  this->byte_pos = 0;

  return 1;
}

static int dvd_plugin_open (input_plugin_t *this_gen, char *mrl) {
  char                *filename;
  dvd_input_plugin_t  *this = (dvd_input_plugin_t *) this_gen;
  int                  ifo_num, title, chapter, to_chapter;

  this->mrl = mrl;

  xprintf (VERBOSE|INPUT, "input_d4d : input_plugin_open >%s<\n", mrl);

  /*
   * do we handle this kind of MRL ?
   */

  if (strncmp (mrl, "d4d://",6))
    return 0;

  filename = (char *) &mrl[6];

  printf ("input_d4d : input_plugin_open media type correct. file name is %s\n",
	  filename);

  if (sscanf (filename, "i%dt%dc%dt%d", &ifo_num, &title, 
	      &chapter, &to_chapter) == 4) {

    printf ("input_d4d: playing title %d from chapter %d -> chapter %d\n",
	    title, chapter, to_chapter);

    return openDVDChapter (this, ifo_num, title, chapter, to_chapter);

  } else {

    sscanf (filename, "VTS_%d_%d.VOB", &this->gVTSMajor, &this->gVTSMinor);

    this->ifo = NULL;

    this->file_lbstart = openDVDFile (this, filename, &this->file_size) ;
    this->file_lbcur = this->file_lbstart;

    if (!this->file_lbstart) {
      printf ("input_d4d: unable to find >%s< on dvd.\n", filename);
      return 0;
    }

    this->file_size_left = this->file_size;
  }
 
  return 1 ;
}

static off_t dvd_plugin_read (input_plugin_t *this_gen, 
			      char *buf, off_t nlen) {
  dvd_input_plugin_t *this = (dvd_input_plugin_t *) this_gen;

  if (nlen != DVD_VIDEO_LB_LEN) {
    printf ("input_d4d: %d bytes is not a sector\n", (int) nlen); 
    return 0;
  }

  if (this->ifo) {

    if ((this->cellp_pos + this->title_start) == this->lb_end)
      return 0;

    if (++this->cellp_pos > this->cell_end) {       /* next cell */
      if (_dvd_next_cell (this) < 0) {
	this->cpos_index++;
	if (_dvd_next_cell (this) < 0)
	  return 0;
      }
    }

    this->byte_pos += DVD_VIDEO_LB_LEN;
    lseek (this->raw_fd, (this->title_start + this->cellp_pos) 
	   * DVD_VIDEO_LB_LEN, SEEK_SET) ;

  } else {

    if (this->file_size_left < nlen)
      return 0;
    
    this->file_lbcur++;
    this->file_size_left -= DVD_VIDEO_LB_LEN;

  }
    
  if (read (this->raw_fd, buf, DVD_VIDEO_LB_LEN)) {

    /* printf ("read to %08x => %02x %02x %02x %02x\n",buf,buf[0],buf[1],buf[2],buf[3]); */
    
    if (this->bEncrypted)
      CSSDescramble (buf, this->key_title);
    
    
    return DVD_VIDEO_LB_LEN;
  } else
    printf ("read error in input_d4d plugin\n");

  return 0;
}

static buf_element_t *dvd_plugin_read_block (input_plugin_t *this_gen, 
					     fifo_buffer_t *fifo, off_t nlen) {
  /* dvd_input_plugin_t *this = (dvd_input_plugin_t *) this_gen; */
  buf_element_t      *buf = fifo->buffer_pool_alloc (fifo);
  int                 num_bytes_read;

  buf->content = buf->mem;
  
  num_bytes_read = dvd_plugin_read (this_gen, buf->mem, nlen);

  if (num_bytes_read < nlen) {
    buf->free_buffer (buf);
    return NULL;
  }

  buf->size = num_bytes_read;
  return buf;
}

static off_t dvd_plugin_seek (input_plugin_t *this_gen, 
			      off_t offset, int origin) {
  dvd_input_plugin_t *this = (dvd_input_plugin_t *) this_gen;

  if (this->ifo) {
    off_t seek_pos = 0;

    switch (origin) {
    case SEEK_END:
      seek_pos = this->file_size - offset;

    case SEEK_SET:
      seek_pos = offset;
      break;

    case SEEK_CUR:

      if (offset==0) {
	return this->byte_pos;
      }

      seek_pos = this->byte_pos + offset;
      break;
    default:
      printf ("error in input_d4d plugin seek:%d is an unknown origin\n", origin);
    }
    
    this->cpos_index     = this->clink[this->chapter]-1;
    this->caddr_index = -1;
    this->cellp_pos = 0;
    this->cell_end = 0;
    this->byte_pos = 0;

    while (this->byte_pos < seek_pos) {
      if (++this->cellp_pos > this->cell_end) {       /* next cell */
	if (_dvd_next_cell (this) < 0) {
	  this->cpos_index++;
	  if (_dvd_next_cell (this) < 0)
	    break;
	}
      }
      this->byte_pos += DVD_VIDEO_LB_LEN;
    }

    return this->byte_pos;
  } else {

    offset /= DVD_VIDEO_LB_LEN;
  
    switch (origin) {
    case SEEK_END:
      offset = (this->file_size / DVD_VIDEO_LB_LEN) - offset;
      
    case SEEK_SET:
      this->file_lbcur = this->file_lbstart + offset;
      this->file_size_left = this->file_size - (offset * DVD_VIDEO_LB_LEN);
      break;
    case SEEK_CUR:
      if (offset) {
	this->file_lbcur += offset;
	this->file_size_left = this->file_size - 
	((this->file_lbcur - this->file_lbstart) * DVD_VIDEO_LB_LEN);
      } else {
	return (this->file_lbcur - this->file_lbstart) * 
	  (off_t) DVD_VIDEO_LB_LEN;
      }
      
      break;
    default:
      fprintf (stderr, "error in input_d4d plugin seek: %d is an unknown "
	       "origin\n", origin);
    }
    
    return lseek (this->raw_fd, 
		  this->file_lbcur * (off_t) DVD_VIDEO_LB_LEN, SEEK_SET) 
      - this->file_lbstart * (off_t) DVD_VIDEO_LB_LEN;
    
  }

  return 0;
}

static off_t dvd_plugin_get_length (input_plugin_t *this_gen) {
  dvd_input_plugin_t *this = (dvd_input_plugin_t *) this_gen;

  return this->file_size;
}

static uint32_t dvd_plugin_get_capabilities (input_plugin_t *this) {
  return INPUT_CAP_SEEKABLE | INPUT_CAP_BLOCK | INPUT_CAP_AUTOPLAY | INPUT_CAP_GET_DIR;
}

static uint32_t dvd_plugin_get_blocksize (input_plugin_t *this_gen) {

  return DVD_VIDEO_LB_LEN;
}

static int dvd_plugin_eject_media (input_plugin_t *this_gen) {
  int   ret, status;
  int   fd;

  if((fd = open(DVD, O_RDONLY|O_NONBLOCK)) > -1) {

#if defined (__linux__)
    if((status = ioctl(fd, CDROM_DRIVE_STATUS, CDSL_CURRENT)) > 0) {
      switch(status) {
      case CDS_TRAY_OPEN:
	if((ret = ioctl(fd, CDROMCLOSETRAY)) != 0) {
	  xprintf(VERBOSE|INPUT, "CDROMCLOSETRAY failed: %s\n", 
		  strerror(errno));  
	}
	break;
      case CDS_DISC_OK:
	if((ret = ioctl(fd, CDROMEJECT)) != 0) {
	  xprintf(VERBOSE|INPUT, "CDROMEJECT failed: %s\n", strerror(errno));  
	}
	break;
      }
    }
    else {
      xprintf(VERBOSE|INPUT, "CDROM_DRIVE_STATUS failed: %s\n", 
	      strerror(errno));
      close(fd);
      return 0;
    }

#elif defined (__NetBSD__) || defined (__OpenBSD__) || defined (__FreeBSD__)

    if (ioctl(fd, CDIOCALLOW) == -1) {
      perror("ioctl(cdromallow)");
    } else {
      if (ioctl(fd, CDIOCEJECT) == -1) {
        perror("ioctl(cdromeject)");
      }
    }

#endif

    close(fd);
  }
  return 1;
}

static void dvd_plugin_close (input_plugin_t *this_gen) {
  dvd_input_plugin_t *this = (dvd_input_plugin_t *) this_gen;

  close_drive(this);

  if (this->ifo) {
    ifoClose (this->ifo);
    this->ifo = NULL;
  }
}

static char *dvd_plugin_get_description (input_plugin_t *this_gen) {

  return "dvd device input plugin, css featured";
}

static char *dvd_plugin_get_identifier (input_plugin_t *this_gen) {

  return "d4d";
}

static mrl_t **dvd_plugin_get_dir (input_plugin_t *this_gen, 
				   char *filename, int *num_entries) {

  dvd_input_plugin_t *this = (dvd_input_plugin_t *) this_gen;
  int                 i, j, fd, num_ifos, ifo;
  int                 num_files;
  ifo_ptt_t          *ptt;
  ifo_ptt_sub_t      *title;

  if (filename) {
    *num_entries = 0;
    return NULL;
  }

  /*
   * first scan disc using ifo files
   */

  reset_dvd_drive ();

  num_files = 0;

  if (open_drive(this)>=0) {
  
    /*
     * parse all IFO files
     */

    num_ifos = get_num_ifos (this);

    for (ifo = 1; ifo<=num_ifos; ifo++) {

      if (parse_ifo (this, ifo)) {

	ptt = ifoGetPTT (this->ifo);

	title     = ptt->title; 

	for(j = 0; j < ptt->num; j++) {
	  for (i = 0; i < title->num; i++) {

	    sprintf (this->mrls[num_files]->mrl, "d4d://i%dt%dc%dt%d", 
		     ifo, j, i, i);
	    this->mrls[num_files]->size = 0;
	    this->mrls[num_files]->type = mrl_dvd;
	    num_files++;

	  }
	  title++;
	}

      }

    }
    close_drive (this);
  }


  /*
   * now append "raw" file entries
   */

  if((fd = open(DVD, O_RDONLY|O_NONBLOCK)) > -1) {

    UDFListDir (fd, "/VIDEO_TS", MAX_DIR_ENTRIES, 
		this->filelist, &this->num_filelist);

    for (i=0; i<this->num_filelist; i++) {
      int nLen;

      nLen = strlen (this->filelist[i]);

      if (nLen<4) 
	continue;
      
      if (!strcasecmp (&this->filelist[i][nLen-4], ".VOB")) {
	char str[1024];

	sprintf (this->mrls[num_files]->mrl, "d4d://%s", this->filelist[i]); 
	this->mrls[num_files]->type = mrl_dvd;

	/* determine size */
	sprintf (str, "/VIDEO_TS/%s", this->filelist[i]);
	UDFFindFile(fd, str, &this->mrls[num_files]->size);

	num_files++;
      }

    }
    close (fd);
  } 

  *num_entries = num_files;

  return this->mrls;
}

static char **dvd_plugin_get_autoplay_list (input_plugin_t *this_gen,
					    int *nFiles) {
  dvd_input_plugin_t *this = (dvd_input_plugin_t *) this_gen;
  int                 i, j=0, alltitles, nfilestmp;
  ifo_ptt_t          *ptt;
  ifo_ptt_sub_t      *title;
  char                buf[256]="";
  int                 movie_ifo;

  reset_dvd_drive ();

  if (open_drive(this)<0) {
    *nFiles = 0;
    return 0;
  }
  
  /*
   * parse IFO files
   */

  movie_ifo = find_movie_ifo(this);

  if (!parse_ifo (this, movie_ifo)) {
    *nFiles = 0;
    return NULL;
  }

  ptt = ifoGetPTT (this->ifo);

  title = ptt->title; /* assuming title 0 is the movie */
  alltitles = ptt->num;
  nfilestmp = 0;

#ifdef ALL_TITLES
  for(j = 0; j < alltitles; j++) {
#endif
    for (i = 0; i < title->num; i++) {
      sprintf (this->filelist[nfilestmp], "d4d://i%dt%dc%dt%d",movie_ifo, j, i, i);
      nfilestmp++;
    }
#ifdef ALL_TITLES    
    title++;
  }
#endif

  *nFiles = nfilestmp;

  close_drive (this);

  return this->filelist;
}

static off_t dvd_plugin_get_current_pos (input_plugin_t *this_gen){
  dvd_input_plugin_t *this = (dvd_input_plugin_t *) this_gen;

  return (this->file_lbcur * DVD_VIDEO_LB_LEN);
}

static char* dvd_plugin_get_mrl (input_plugin_t *this_gen) {
  dvd_input_plugin_t *this = (dvd_input_plugin_t *) this_gen;

  return this->mrl;
}

static int dvd_plugin_get_optional_data (input_plugin_t *this_gen, 
					 void *data, int data_type) {
  dvd_input_plugin_t *this = (dvd_input_plugin_t *) this_gen;
  void *ptr;

  switch(data_type) {

  case INPUT_OPTIONAL_DATA_CLUT:
    ptr = (void *)ifoGetCLUT(this->pgci);
	
    if(ptr) {
      memcpy(data, ptr, sizeof(clut_t)*16); 
      return INPUT_OPTIONAL_SUCCESS;
    } else {
      return INPUT_OPTIONAL_UNSUPPORTED;
    }
    break;
    /*    
  case INPUT_OPTIONAL_DATA_AUDIOLANG:
    return INPUT_OPTIONAL_SUCCESS;
    break;
    */
    /*
  case INPUT_OPTIONAL_DATA_LANGCHAN: {
    char *desired_chan;
    int  *channum;

    if(((char*) data) == NULL)
      return INPUT_OPTIONAL_UNSUPPORTED;

    desired_chan = strdup(((char*) data));
    channum = ifoGETAudioTrack(desired_chan);
    if(channum > 0) {
      data = (void *) channum;
      return INPUT_OPTIONAL_SUCCESS;
    }
    else {
      return INPUT_OPTIONAL_UNSUPPORTED;
    }	
  }
  break;
    */
  }

  return INPUT_OPTIONAL_UNSUPPORTED;
}

static int dvd_plugin_is_branch_possible (input_plugin_t *this_gen, char *next_mrl) {
  
  dvd_input_plugin_t *this = (dvd_input_plugin_t *) this_gen;
  char               *filename;
  int                 ifo_num, title, chapter, to_chapter;

  printf ("input_dvd: is_branch_possible to %s ?\n", next_mrl);

  /*
   * do we handle this kind of MRL ?
   */

  if (strncmp (next_mrl, "d4d://",6))
    return 0;

  filename = (char *) &next_mrl[6];

  if (sscanf (filename, "i%dt%dc%dt%d", &ifo_num, &title, 
	      &chapter, &to_chapter) == 4) {
      
    if ((this->title == title) 
        && (this->to_chapter == (chapter-1))) {
      printf ("input_d4d: branching is possible\n");
      return 1;
    }
  }
  return 0;
}

input_plugin_t *init_input_plugin (int iface, config_values_t *config) {

  dvd_input_plugin_t *this;
  int i, fd;

  xine_debug = config->lookup_int (config, "xine_debug", 0);
  
  
  if (iface != 2) {
    printf ("input_d4d: input plugin doesn't support plugin API version %d.\n",
            iface);
    return NULL;
  }

  this = (dvd_input_plugin_t *) malloc (sizeof (dvd_input_plugin_t));
  
  for (i = 0; i < MAX_DIR_ENTRIES; i++) {
    this->filelist[i]       = (char *) malloc (256);
    this->filelist2[i]      = (char *) malloc (256);
    this->mrls[i]           = (mrl_t *) malloc(sizeof(mrl_t));
    this->mrls[i]->mrl      = (char *) malloc (256);
  }
  
  this->mrls_allocated_entries = MAX_DIR_ENTRIES;
  
  this->input_plugin.interface_version = INPUT_PLUGIN_IFACE_VERSION;
  this->input_plugin.get_capabilities  = dvd_plugin_get_capabilities;
  this->input_plugin.open              = dvd_plugin_open;
  this->input_plugin.read              = dvd_plugin_read;
  this->input_plugin.read_block        = dvd_plugin_read_block;
  this->input_plugin.seek              = dvd_plugin_seek;
  this->input_plugin.get_current_pos   = dvd_plugin_get_current_pos;
  this->input_plugin.get_length        = dvd_plugin_get_length;
  this->input_plugin.get_blocksize     = dvd_plugin_get_blocksize;
  this->input_plugin.eject_media       = dvd_plugin_eject_media;
  this->input_plugin.close             = dvd_plugin_close;
  this->input_plugin.get_identifier    = dvd_plugin_get_identifier;
  this->input_plugin.get_description   = dvd_plugin_get_description;
  this->input_plugin.get_dir           = dvd_plugin_get_dir;
  this->input_plugin.get_mrl           = dvd_plugin_get_mrl;
  this->input_plugin.get_autoplay_list = dvd_plugin_get_autoplay_list;
  this->input_plugin.get_optional_data = dvd_plugin_get_optional_data;
  this->input_plugin.handle_input_event= NULL;
  this->input_plugin.is_branch_possible= dvd_plugin_is_branch_possible;
  
  this->dvd_fd  = -1;
  this->raw_fd  = -1;
  this->mrl     = NULL;
  this->config  = config;
  this->ifo     = NULL;
  
  /*
   * find out if raw device is available
   */
  fd = open(RDVD, O_RDONLY | O_NONBLOCK);
  if (fd < 0) {
    printf ("input_d4d: unable to open raw dvd drive (%s): %s\n", 
	    RDVD, strerror(errno));
    
  } else {
    printf ("input_d4d: using raw device %s\n", RDVD);
    close (fd);
  }
  
  return (input_plugin_t *) this;
}




